"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Parser = void 0;
/* istanbul ignore file */
const stream = __importStar(require("stream"));
const consts = __importStar(require("./consts"));
const debug_1 = __importDefault(require("debug"));
const frame_1 = __importDefault(require("./frame"));
const debug = (0, debug_1.default)('zigbee-herdsman:adapter:ezsp:uart');
class Parser extends stream.Transform {
    buffer;
    flagXONXOFF;
    constructor(flagXONXOFF = false) {
        super();
        this.flagXONXOFF = flagXONXOFF;
        this.buffer = Buffer.from([]);
    }
    _transform(chunk, _, cb) {
        if (this.flagXONXOFF && (chunk.indexOf(consts.XON) >= 0 || chunk.indexOf(consts.XOFF) >= 0)) {
            // XXX: should really throw, but just assert for now to flag potential problematic setups
            console.assert(false, `Host driver did not remove XON/XOFF from input stream. Driver not setup for XON/XOFF?`);
        }
        if (chunk.indexOf(consts.CANCEL) >= 0) {
            this.buffer = Buffer.from([]);
            chunk = chunk.subarray(chunk.lastIndexOf(consts.CANCEL) + 1);
        }
        if (chunk.indexOf(consts.SUBSTITUTE) >= 0) {
            this.buffer = Buffer.from([]);
            chunk = chunk.subarray(chunk.indexOf(consts.FLAG) + 1);
        }
        debug(`<-- [${chunk.toString('hex')}]`);
        this.buffer = Buffer.concat([this.buffer, chunk]);
        this.parseNext();
        cb();
    }
    parseNext() {
        if (this.buffer.length) {
            const place = this.buffer.indexOf(consts.FLAG);
            if (place >= 0) {
                const frameLength = place + 1;
                if (this.buffer.length >= frameLength) {
                    const frameBuffer = this.unstuff(this.buffer.subarray(0, frameLength));
                    try {
                        const frame = frame_1.default.fromBuffer(frameBuffer);
                        if (frame) {
                            debug(`--> parsed ${frame}`);
                            this.emit('parsed', frame);
                        }
                    }
                    catch (error) {
                        debug(`--> error ${error.stack}`);
                    }
                    this.buffer = this.buffer.subarray(frameLength);
                    this.parseNext();
                }
            }
        }
    }
    unstuff(s) {
        /* Unstuff (unescape) a string after receipt */
        let escaped = false;
        const out = Buffer.alloc(s.length);
        let outIdx = 0;
        for (let idx = 0; idx < s.length; idx += 1) {
            const c = s[idx];
            if (escaped) {
                out.writeUInt8(c ^ consts.STUFF, outIdx++);
                escaped = false;
            }
            else {
                if (c === consts.ESCAPE) {
                    escaped = true;
                }
                else if (c === consts.XOFF || c === consts.XON) {
                    // skip
                }
                else {
                    out.writeUInt8(c, outIdx++);
                }
            }
        }
        return out.subarray(0, outIdx);
    }
    reset() {
        // clear buffer
        this.buffer = Buffer.from([]);
    }
}
exports.Parser = Parser;
//# sourceMappingURL=parser.js.map